==> pcv3cloudconnector-nodejs-1.0.13.tgz: package/src/proficloud_connector.js <==
var https = require('https');
var mqtt = require('mqtt');
var fs = require('fs');


const ebjca_api = "/middleware/ejbca/api/";
const command_topic_regex = /kp1\/proficloud-v1\/cex\/[\w\/]*command\/[\w]*\/status/;


/**
 * @description This class represents a virtual device of the Proficloud.io . It includes the hole connection management and the data encoding.
 * @param {string} uuid 
 */
function Connector(uuid, proficloud_api="env.kaa.proficloud-production.io", proficloud_mqtt_url="mqtts://mqtt.proficloud-production.io:8883") {
    var self = this;
    this.uuid = uuid;
    this.connecting = false;
    this.authBundle = "";
    this.client = null; // mqtt client object
    this.connectCallback = function () { };
    this.connectionLostCallback = function () { };
    this.commandCallback = function () { };
    this.proficloud_api = proficloud_api;
    this.proficloud_mqtt_url = proficloud_mqtt_url

    try {
        this.authBundle = JSON.parse(fs.readFileSync(__dirname + '/certs/' + uuid + '_auth_bundle.json'));
    } catch (e) {
        console.log("No authBundle for UUID " + this.uuid + " stored");
    }

    /**
     * @description This function start the connection process and sets the required callbacks for the client
     * @param {function} connectCallback 
     * @param {function} connectionLostCallback 
     */
    this.connect = function (connectCallback, connectionLostCallback) {
        // prevents problems when starting the conneciton process twice
        if (self.connecting) {
            return;
        }
        self.connecting = true;
        self.getCertState();
        if (connectCallback != undefined) { // use default callbacks if the parameter is undefined
            self.connectCallback = connectCallback;
        }
        if (connectionLostCallback != undefined) { // use default callbacks if the parameter is undefined
            self.connectionLostCallback = connectionLostCallback;
        }
    };

    /**
     * @description This function disconnects the mqtt client
     */
    this.disconnect = function () {
        console.log("disconnect mqtt client");
        try { // catch if client was not initialized or connected
            self.client.end(true);
        } catch (e){
            console.log("Disconnect called, but client was not connected");
        }
    };

    /**
     * @description This function sends TimeSeriesData to the Proficloud via MQTT
     * @param {Object} tsd_data 
     */
    this.sendTSD = function (tsd_data) {
        var tsd_topic = "kp1/proficloud-v1/dcx/" + this.authBundle.DeviceToken + "/json/" + Math.floor(Math.random() * Math.floor(1000));
        if (this.client != undefined) {
            if (Array.isArray(tsd_data)) { // check if data is buffer or single data batch
                this.client.publish(tsd_topic, JSON.stringify(tsd_data));
            } else {
                this.client.publish(tsd_topic, JSON.stringify([tsd_data]));
            }
        } else {
            throw "error in sendTSD";
        }
    };

    /**
     * @description This function sends TrafficLight to the Proficloud via MQTT
     * @param {Object} traffic_light_data 
     */
    this.sendTrafficLight = function (traffic_light_data) {
        var tsd_topic = "kp1/proficloud-v1/dcx/" + this.authBundle.DeviceToken + "/json/" + Math.floor(Math.random() * Math.floor(1000));
        if (this.client != undefined && traffic_light_data.color != undefined && traffic_light_data.msg != undefined) {
            this.client.publish(tsd_topic, JSON.stringify([{ "con_TrafficLightColor": parseInt(traffic_light_data.color), "con_TrafficLightMessage": traffic_light_data.msg }]));
        } else {
            throw "error in sendTrafficLight";
        }
    };

    /**
     * @description This function sends logs to the Proficloud via MQTT
     * @param {Object} log_data 
     */
    this.sendLog = function (log_data) {
        var tsd_topic = "kp1/proficloud-v1/dcx/" + this.authBundle.DeviceToken + "/json/" + Math.floor(Math.random() * Math.floor(1000));
        if (this.client != undefined && log_data.level != undefined && log_data.tag != undefined && log_data.msg != undefined) {
            this.client.publish(tsd_topic, JSON.stringify([{ "con_LogLevel": parseInt(log_data.level), "con_LogTags": log_data.tag, "con_LogMessage": log_data.msg }]));
        } else {
            throw "error in sendLog"
        }
    };

    /**
     * @description This function sends metadata to the Proficloud via MQTT
     * @param {Object} meta_data 
     */
    this.sendMeta = function (meta_data) {
        var meta_topic = "kp1/proficloud-v1/epmx/" + this.authBundle.DeviceToken + "/update/keys/" + Math.floor(Math.random() * Math.floor(1000));
        if (this.client != undefined) {
            this.client.publish(meta_topic, JSON.stringify(meta_data));
        } else {
            throw "error in sendMeta"
        }
    }

    /**
     * @description This function connects the mqtt client with the proficloud
     */
    this.connectMQTT = function () {
        var mqtt_options = {
            key: this.authBundle.DeviceCertKey,
            cert: this.authBundle.DeviceCert,
            ca: this.authBundle.CACert,
            rejectUnauthorized: false, // TODO wrong hostname in cert
            keepalive: 10
        };
        if (this.authBundle.DeviceCertKey == undefined || this.authBundle.DeviceCert == undefined || this.authBundle.CACert == undefined) {
            self.connecting = false;
            self.connectionLostCallback();
            return;
        }
        self.client = mqtt.connect(proficloud_mqtt_url, mqtt_options);
        self.client.on("connect", function () {
            self.connecting = false;
            console.log("connected");
            self.connectCallback();
        });
        this.client.on("message", this.incomingMessage);

        this.client.on("error", function (error) {
            self.connecting = false;
            console.log("Can't connect" + error);
            self.connectionLostCallback();
        });
    };

    /**
     * @description This functions sends an acceptance call to the Proficloud after the successful transfer of the auth bundle
     */
    this.acceptCert = function () {
        var options = {
            hostname: proficloud_api,
            port: 443,
            path: ebjca_api + "certificate/device/acceptance/" + uuid,
            method: 'PUT'
        };
        https.get(options, (resp) => {
            let data = '';
            resp.on('data', (chunk) => {
                data += chunk;
            });

            resp.on('end', () => {
                console.log("new certificates accepted");
            });

        }).on("error", (err) => {
            console.error("Error: " + err.message);
        });
    }

    /**
     * @description This function requests the Proficloud and checks wheter there is a new auth bundle available
     */
    this.getCertState = function () {
        https.get("https://" + proficloud_api + ebjca_api + "certificate/status/" + this.uuid, (resp) => {
            let data = '';
            resp.on('data', (chunk) => {
                data += chunk;
            });

            resp.on('end', () => {
                const dataObject = JSON.parse(data);
                if (dataObject.generated) {
                    if (dataObject.ExpiryDate) {
                        if (this.authBundle.DeviceToken != undefined) {
                            console.log("The client certificate is valid, but will expire soon; Requesting a renewed");
                            this.renewCertificate();
                        } else {
                            self.connecting = false;
                            console.error("Error: The client certificate is valid, but will expire soon; No device token set!")
                        }
                    } else {
                        this.connectMQTT();
                    }
                } else {
                    this.getNewCertificate();
                }
            });

        }).on("error", (err) => {
            self.connecting = false;
            console.error("Error: " + err.message);
        });
    };

    /**
     * @description This function request a renewed certificate from the Proficloud
     */
    this.renewCertificate = function () {
        var options = {
            hostname: proficloud_api,
            port: 443,
            path: ebjca_api + "certificate/device/" + uuid,
            method: 'GET'
        };
        https.get(options, (resp) => {
            let data = '';
            resp.on('data', (chunk) => {
                data += chunk;
            });

            resp.on('end', () => {
                dataObject = JSON.parse(data);
                this.authBundle.DeviceCert = dataObject.DeviceCert;
                this.authBundle.DeviceCertKey = dataObject.DeviceCertKey;

                this.storeAuthBundle(() => {
                    self.acceptCert();
                    self.connectMQTT();
                });

            });

        }).on("error", (err) => {
            self.connecting = false;
            console.error("Error: " + err.message);
        });
    };

    /**
     * @description This function stores the authBundle of the object in a file
     */
    this.storeAuthBundle = function(callback) {
        fs.promises.mkdir(__dirname + '/certs', { recursive: true }).catch(console.error);
        fs.writeFile(__dirname + '/certs/' + uuid + '_auth_bundle.json', JSON.stringify(this.authBundle), (err) => {
            if (err) {
                throw err;
            }
            console.log("authBundle is saved.");
            callback();
        });
    }

    /**
     * @description This function request a new auth bundle from the Proficloud
     */
    this.getNewCertificate = function () {
        var options = {
            hostname: proficloud_api,
            port: 443,
            path: ebjca_api + "certificate/device/" + uuid,
            method: 'PUT'
        };
        https.get(options, (resp) => {
            let data = '';
            resp.on('data', (chunk) => {
                data += chunk;
            });

            resp.on('end', () => {
                this.authBundle = JSON.parse(data);
                this.storeAuthBundle(() => {
                    self.acceptCert();
                    self.connectMQTT();
                });

            });

        }).on("error", (err) => {
            self.connecting = false;
            console.error("Error: " + err.message);
        });
    };

    /**
     * @description This function registers the callback which will be called if a
     * @param {function} callback 
     */
    this.registerCommandCallback = function(callback) {
        this.commandCallback = callback;
    };

    /**
     * @description This function handles the incoming messages from the mqtt and forwards them to the registered callbacks
     * @param {String} mqtt topic of the received message
     * @param {Buffer} payload of the received message
     */
    this.incomingMessage = function (topic , msg) {

        if (topic.match(command_topic_regex)){
            const command = topic.split("/")[5];
            const command_result_topic = "kp1/proficloud-v1/cex/" + self.authBundle.DeviceToken + "/result/" + command;
            commands_object = JSON.parse(msg.toString());
            for (index in commands_object){
                const command_object = commands_object[index];

                // publish result to the proficloud as acceptance message
                self.client.publish(command_result_topic, JSON.stringify([{"id": command_object["id"], "statusCode": 200, "payload": "command will be executed"}]));

                try {
                    self.commandCallback(command, command_object["payload"]);
                } catch (e){
                    console.error("Error during calling remote command callback: " + e);
                }
            }
            
        }
    };
}

module.exports.Connector = Connector;

==> pcv3cloudconnector-nodejs-1.0.14.tgz: package/src/proficloud_connector.js <==
var https = require('https');
var mqtt = require('mqtt');
var fs = require('fs');


const ebjca_api = "/middleware/ejbca/api/";
const command_topic_regex = /kp1\/proficloud-v1\/cex\/[\w\/]*command\/[\w]*\/status/;


/**
 * @description This class represents a virtual device of the Proficloud.io . It includes the hole connection management and the data encoding.
 * @param {string} uuid 
 */
function Connector(uuid, proficloud_api = "env.kaa.proficloud-production.io", proficloud_mqtt_url = "mqtts://mqtt.proficloud-production.io:8883") {
    var self = this;
    this.uuid = uuid;
    this.connecting = false;
    this.authBundle = "";
    this.client = null; // mqtt client object
    this.connectCallback = function () { };
    this.connectionLostCallback = function () { };
    this.commandCallback = function () { };
    this.proficloud_api = proficloud_api;
    this.proficloud_mqtt_url = proficloud_mqtt_url;
    this.tsd_buffer = [];

    try {
        this.authBundle = JSON.parse(fs.readFileSync(__dirname + '/certs/' + uuid + '_auth_bundle.json'));
    } catch (e) {
        console.log("No authBundle for UUID " + this.uuid + " stored");
    }

    /**
     * @description This function start the connection process and sets the required callbacks for the client
     * @param {function} connectCallback 
     * @param {function} connectionLostCallback 
     */
    this.connect = function (connectCallback, connectionLostCallback) {
        // prevents problems when starting the conneciton process twice
        if (self.connecting) {
            return;
        }
        self.connecting = true;
        self.getCertState();
        if (connectCallback != undefined) { // use default callbacks if the parameter is undefined
            self.connectCallback = connectCallback;
        }
        if (connectionLostCallback != undefined) { // use default callbacks if the parameter is undefined
            self.connectionLostCallback = connectionLostCallback;
        }
    };

    /**
     * @description This function disconnects the mqtt client
     */
    this.disconnect = function () {
        console.log("disconnect mqtt client");
        try { // catch if client was not initialized or connected
            self.client.end(true);
        } catch (e) {
            console.log("Disconnect called, but client was not connected");
        }
    };

    /**
     * @description This function sends TimeSeriesData to the Proficloud via MQTT
     * @param {Object} tsd_data 
     */
    this.sendTSD = function (tsd_data) {
        var tsd_topic = "kp1/proficloud-v1/dcx/" + this.authBundle.DeviceToken + "/json/" + Math.floor(Math.random() * Math.floor(1000));
        if (this.client != undefined) {
            if (Array.isArray(tsd_data)) { // check if data is buffer or single data batch
                this.client.publish(tsd_topic, JSON.stringify(tsd_data));
            } else {
                this.client.publish(tsd_topic, JSON.stringify([tsd_data]));
            }
        } else {
            throw "error in sendTSD";
        }
    };

    /**
    * @description This function adds the tsd_data object to the buffer array. Which will be sent frequently when activated
    * @param {Object} tsd_data 
    */
    this.sendTSDToBuffer = function (tsd_data) {
        this.tsd_buffer.push(tsd_data);
    };

    /**
    * @description This function configures, starts and stops the TSD buffer send interval
    * @param {bool} activated true if the interval should start
    * @param {number} interaval in milliseconds for sending TSD buffer
    */
    this.setTSDBufferConfig = function (activated, interval) {
        if (activated) {
            this.tsd_interval = setInterval(() => {
                if (this.tsd_buffer.length != 0){
                    this.sendTSD(this.tsd_buffer);
                    this.tsd_buffer = [];
                }
            }, interval)
        } else if (this.tsd_interval) {
            clearInterval(this.tsdInterval);
        }
    };

    /**
     * @description This function sends TrafficLight to the Proficloud via MQTT
     * @param {Object} traffic_light_data 
     */
    this.sendTrafficLight = function (traffic_light_data) {
        var tsd_topic = "kp1/proficloud-v1/dcx/" + this.authBundle.DeviceToken + "/json/" + Math.floor(Math.random() * Math.floor(1000));
        if (this.client != undefined && traffic_light_data.color != undefined && traffic_light_data.msg != undefined) {
            this.client.publish(tsd_topic, JSON.stringify([{ "con_TrafficLightColor": parseInt(traffic_light_data.color), "con_TrafficLightMessage": traffic_light_data.msg }]));
        } else {
            throw "error in sendTrafficLight";
        }
    };

    /**
     * @description This function sends logs to the Proficloud via MQTT
     * @param {Object} log_data 
     */
    this.sendLog = function (log_data) {
        var tsd_topic = "kp1/proficloud-v1/dcx/" + this.authBundle.DeviceToken + "/json/" + Math.floor(Math.random() * Math.floor(1000));
        if (this.client != undefined && log_data.level != undefined && log_data.tag != undefined && log_data.msg != undefined) {
            this.client.publish(tsd_topic, JSON.stringify([{ "con_LogLevel": parseInt(log_data.level), "con_LogTags": log_data.tag, "con_LogMessage": log_data.msg }]));
        } else {
            throw "error in sendLog"
        }
    };

    /**
     * @description This function sends metadata to the Proficloud via MQTT
     * @param {Object} meta_data 
     */
    this.sendMeta = function (meta_data) {
        var meta_topic = "kp1/proficloud-v1/epmx/" + this.authBundle.DeviceToken + "/update/keys/" + Math.floor(Math.random() * Math.floor(1000));
        if (this.client != undefined) {
            this.client.publish(meta_topic, JSON.stringify(meta_data));
        } else {
            throw "error in sendMeta"
        }
    }

    /**
     * @description This function connects the mqtt client with the proficloud
     */
    this.connectMQTT = function () {
        var mqtt_options = {
            key: this.authBundle.DeviceCertKey,
            cert: this.authBundle.DeviceCert,
            ca: this.authBundle.CACert,
            rejectUnauthorized: false, // TODO wrong hostname in cert
            keepalive: 10
        };
        if (this.authBundle.DeviceCertKey == undefined || this.authBundle.DeviceCert == undefined || this.authBundle.CACert == undefined) {
            self.connecting = false;
            self.connectionLostCallback();
            return;
        }
        self.client = mqtt.connect(proficloud_mqtt_url, mqtt_options);
        self.client.on("connect", function () {
            self.connecting = false;
            console.log("connected");
            self.connectCallback();
        });
        this.client.on("message", this.incomingMessage);

        this.client.on("error", function (error) {
            self.connecting = false;
            console.log("Can't connect" + error);
            self.connectionLostCallback();
        });
    };

    /**
     * @description This functions sends an acceptance call to the Proficloud after the successful transfer of the auth bundle
     */
    this.acceptCert = function () {
        var options = {
            hostname: proficloud_api,
            port: 443,
            path: ebjca_api + "certificate/device/acceptance/" + uuid,
            method: 'PUT'
        };
        https.get(options, (resp) => {
            let data = '';
            resp.on('data', (chunk) => {
                data += chunk;
            });

            resp.on('end', () => {
                console.log("new certificates accepted");
            });

        }).on("error", (err) => {
            console.error("Error: " + err.message);
        });
    }

    /**
     * @description This function requests the Proficloud and checks wheter there is a new auth bundle available
     */
    this.getCertState = function () {
        https.get("https://" + proficloud_api + ebjca_api + "certificate/status/" + this.uuid, (resp) => {
            let data = '';
            resp.on('data', (chunk) => {
                data += chunk;
            });

            resp.on('end', () => {
                const dataObject = JSON.parse(data);
                if (dataObject.generated) {
                    if (dataObject.ExpiryDate) {
                        if (this.authBundle.DeviceToken != undefined) {
                            console.log("The client certificate is valid, but will expire soon; Requesting a renewed");
                            this.renewCertificate();
                        } else {
                            self.connecting = false;
                            console.error("Error: The client certificate is valid, but will expire soon; No device token set!")
                        }
                    } else {
                        this.connectMQTT();
                    }
                } else {
                    this.getNewCertificate();
                }
            });

        }).on("error", (err) => {
            self.connecting = false;
            console.error("Error: " + err.message);
        });
    };

    /**
     * @description This function request a renewed certificate from the Proficloud
     */
    this.renewCertificate = function () {
        var options = {
            hostname: proficloud_api,
            port: 443,
            path: ebjca_api + "certificate/device/" + uuid,
            method: 'GET'
        };
        https.get(options, (resp) => {
            let data = '';
            resp.on('data', (chunk) => {
                data += chunk;
            });

            resp.on('end', () => {
                dataObject = JSON.parse(data);
                this.authBundle.DeviceCert = dataObject.DeviceCert;
                this.authBundle.DeviceCertKey = dataObject.DeviceCertKey;

                this.storeAuthBundle(() => {
                    self.acceptCert();
                    self.connectMQTT();
                });

            });

        }).on("error", (err) => {
            self.connecting = false;
            console.error("Error: " + err.message);
        });
    };

    /**
     * @description This function stores the authBundle of the object in a file
     */
    this.storeAuthBundle = function (callback) {
        fs.promises.mkdir(__dirname + '/certs', { recursive: true }).catch(console.error);
        fs.writeFile(__dirname + '/certs/' + uuid + '_auth_bundle.json', JSON.stringify(this.authBundle), (err) => {
            if (err) {
                throw err;
            }
            console.log("authBundle is saved.");
            callback();
        });
    }

    /**
     * @description This function request a new auth bundle from the Proficloud
     */
    this.getNewCertificate = function () {
        var options = {
            hostname: proficloud_api,
            port: 443,
            path: ebjca_api + "certificate/device/" + uuid,
            method: 'PUT'
        };
        https.get(options, (resp) => {
            let data = '';
            resp.on('data', (chunk) => {
                data += chunk;
            });

            resp.on('end', () => {
                this.authBundle = JSON.parse(data);
                this.storeAuthBundle(() => {
                    self.acceptCert();
                    self.connectMQTT();
                });

            });

        }).on("error", (err) => {
            self.connecting = false;
            console.error("Error: " + err.message);
        });
    };

    /**
     * @description This function registers the callback which will be called if a
     * @param {function} callback 
     */
    this.registerCommandCallback = function (callback) {
        this.commandCallback = callback;
    };

    /**
     * @description This function handles the incoming messages from the mqtt and forwards them to the registered callbacks
     * @param {String} mqtt topic of the received message
     * @param {Buffer} payload of the received message
     */
    this.incomingMessage = function (topic, msg) {

        if (topic.match(command_topic_regex)) {
            const command = topic.split("/")[5];
            const command_result_topic = "kp1/proficloud-v1/cex/" + self.authBundle.DeviceToken + "/result/" + command;
            commands_object = JSON.parse(msg.toString());
            for (index in commands_object) {
                const command_object = commands_object[index];

                // publish result to the proficloud as acceptance message
                self.client.publish(command_result_topic, JSON.stringify([{ "id": command_object["id"], "statusCode": 200, "payload": "command will be executed" }]));

                try {
                    self.commandCallback(command, command_object["payload"]);
                } catch (e) {
                    console.error("Error during calling remote command callback: " + e);
                }
            }

        }
    };
}

module.exports.Connector = Connector;
